/* 
 * C++ implementation of algorithm in "Spatially Adaptive
 * Bayesian P-Splines with Heteroscedastic Errors" by
 * Crainiceanu, Ruppert, and Carroll, based on the Fortran
 * implementation written by Adarsh Joshi.
 *
 * Input via command line: observation values file; 
 * range values file; knot values file;
 * knots_b values file; initial values for sigma2B, sigma2Epsilon, thetaSubB,
 * thetaSubEpsilon, and thetaSubX; number of burn-in iterations; 
 * number of non-burn-in iterations
 *
 * Output to file: sample thetaSubX, sample thetaSubEpsilon,
 * sample thetaSubB, sample sigma2C, sample sigma2D, 
 * sample sigma2Epsilon, sample sigma2B, acceptance ratios for steps 6 and 7,
 * and a file containing
 * calculated sample means and sample standard deviations for the posterior
 * distribution
 */

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cmath>
#include <string>

using namespace std;

#include "gsl/gsl_vector.h"
#include "gsl/gsl_matrix.h"
#include "gsl/gsl_blas.h"
#include "gsl/gsl_rng.h"
#include "gsl/gsl_randist.h"
#include "gsl/gsl_linalg.h"
#include "gsl/gsl_math.h"

void multiplyByDiagonalInverseOnRight (gsl_matrix *A, 
				       gsl_vector *Diag);

void multinvpdsymm_left (gsl_matrix *A, gsl_matrix *B, gsl_matrix *C);

void multinvpdsymm_left (gsl_matrix *A, gsl_vector *B, gsl_vector *C);

void multinvupper_left (gsl_matrix *U, gsl_vector *A);

ofstream f;
ifstream inFile;

int main (int argc, char *argv []) {

  /* Time the total execution. */
  /* The time_t data type is defined in the <ctime> header file
   * and is usually an long int; it makes calculating the time
   * easier with the difftime () function, also in <ctime>. */
  time_t runTimeStart;
  /* Start timing by assigning current time to runTimeStart: */
  time (&runTimeStart);

  /* Data members and initializations: */
  
  /* iter1, iter2, mainIter allow loops to
   * avoid repeated declarations. saveRow keeps track
   * of which iterations to save when thinning the chains. */
  int iter1 = 0;
  int iter2 = 0;
  int mainIter = 0;
  int saveRow = 0;


  /* tempNumberN  holds intermediate calculations. */
  double tempDouble1 = 0.0;
  double tempDouble2 = 0.0;

  /* buffer temporarily stores input from files. */
  char buffer[256];

  /* Read in arguments from command line and initialize
   * constants relying on input: */
  char *yValuesFile = argv [1];
  char *rangeValuesFile = argv [2];
  char *knotsValuesFile = argv [3];
  char *knotsBValuesFile = argv [4];
  char *sigma2BValuesFile = argv [5];
  char *sigma2EpsilonValuesFile = argv [6];
  char *thetaSubBValuesFile = argv [7];
  char *thetaSubEpsilonValuesFile = argv [8];
  char *thetaSubXValuesFile = argv [9];
  double sigma2U = atof (argv [10]);
  double sigma2V = atof (argv [11]);
  double proposalVariance = atof (argv [12]);
  int numberOfBurnInIterations = atoi (argv [13]);
  int numberOfIterations = atoi (argv [14]);
  int numberOfIterationsDividedByFive = numberOfIterations / 5;
  int numberOfKnotValues = 0;
  int numberOfKnotValuesPlusThree = 0;
  int numberOfKnotBValues = 0;
  int numberOfYValues = 0;
  int numberOfThetaSubBValues = 0;


  /* Read in the data. */
  /* First determine how many data values there are: */
  inFile.open (knotsValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << knotsValuesFile << ". " << endl;
  } // if ()
  numberOfKnotValues = 0;
  while ((inFile.peek() != '\n') && (! inFile.eof ())) {
    inFile.getline (buffer, 256);
    ++numberOfKnotValues;
  } // while ()
  inFile.close ();
  inFile.clear ();

  gsl_vector *tempVectorLengthKnots = gsl_vector_alloc (numberOfKnotValues);
  inFile.open (knotsValuesFile, ios::in);
  for (iter1 = 0; iter1 < numberOfKnotValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (tempVectorLengthKnots, iter1, atof (buffer));
  } // for (iter1)
  inFile.close ();
  inFile.clear ();
  numberOfKnotValuesPlusThree = numberOfKnotValues + 3;

  inFile.open (knotsBValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << knotsBValuesFile << ". " << endl;
  } // if ()
  numberOfKnotBValues = 0;
  while ((inFile.peek() != '\n') && (! inFile.eof ())) {
    inFile.getline (buffer, 256);
    ++numberOfKnotBValues;
  } // while ()
  inFile.close ();
  inFile.clear ();

  gsl_vector *knotsB = gsl_vector_alloc (numberOfKnotBValues);
  inFile.open (knotsBValuesFile, ios::in);
  for (iter1 = 0; iter1 < numberOfKnotBValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (knotsB, iter1, atof (buffer));
  } // for (iter1)
  inFile.close ();
  inFile.clear ();

  inFile.open (yValuesFile, ios::in);
  if (! inFile.is_open ()) {

    cout << "Cannot open " << yValuesFile << ". " << endl;
  } // if ()
  numberOfYValues = 0;
  while ((inFile.peek() != '\n') && (! inFile.eof ())) {
    inFile.getline (buffer, 256);
    ++numberOfYValues;
  } // while ()
  inFile.close ();
  inFile.clear ();

  gsl_vector *y = gsl_vector_alloc (numberOfYValues);
  inFile.open (yValuesFile, ios::in);
  for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (y, iter1, atof (buffer));
  } // for (iter1)
  inFile.close ();
  inFile.clear ();
  
  inFile.open (rangeValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << rangeValuesFile << ". " << endl;
  } // if ()
  gsl_vector *tempVectorLengthY = gsl_vector_alloc (numberOfYValues);
  for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (tempVectorLengthY, iter1, atof (buffer));
  } // for (iter1)
  inFile.close (); 
  inFile.clear ();

  inFile.open (sigma2BValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << sigma2BValuesFile << ". " << endl;
  } // if ()
  gsl_vector *sigma2B = gsl_vector_alloc (numberOfKnotValues);
  for (iter1 = 0; iter1 < numberOfKnotValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (sigma2B, iter1, atof (buffer));
  } // for (iter1)
  inFile.close (); 
  inFile.clear ();

  inFile.open (sigma2EpsilonValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << sigma2EpsilonValuesFile << ". " << endl;
  } // if ()
  gsl_vector *sigma2Epsilon = gsl_vector_alloc (numberOfYValues);
  for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (sigma2Epsilon, iter1, atof (buffer));
  } // for (iter1)
  inFile.close (); 
  inFile.clear ();

  inFile.open (thetaSubBValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << thetaSubBValuesFile << ". " << endl;
  } // if ()
  numberOfThetaSubBValues = 0;
  while ((inFile.peek() != '\n') && (! inFile.eof ())) {
    inFile.getline (buffer, 256);
    ++numberOfThetaSubBValues;
  } // while ()
  inFile.close ();
  inFile.clear ();

  inFile.open (thetaSubBValuesFile, ios::in);
  gsl_vector *thetaSubB = gsl_vector_calloc (numberOfThetaSubBValues);
  for (iter1 = 0; iter1 < numberOfThetaSubBValues; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (thetaSubB, iter1, atof (buffer));
  } // for (iter1)
  inFile.close (); 
  inFile.clear ();

  inFile.open (thetaSubEpsilonValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << thetaSubEpsilonValuesFile << ". " << endl;
  } // if ()
  gsl_vector *thetaSubEpsilon = gsl_vector_calloc (numberOfKnotValuesPlusThree);
  for (iter1 = 0; iter1 < numberOfKnotValuesPlusThree; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (thetaSubEpsilon, iter1, atof (buffer));
  } // for (iter1)
  inFile.close (); 
  inFile.clear ();

  inFile.open (thetaSubXValuesFile, ios::in);
  if (! inFile.is_open ()) {
    cout << "Cannot open " << thetaSubXValuesFile << ". " << endl;
  } // if ()
  gsl_vector *thetaSubX = gsl_vector_calloc (numberOfKnotValuesPlusThree);
  for (iter1 = 0; iter1 < numberOfKnotValuesPlusThree; ++iter1) {
    inFile.getline (buffer, 256);
    gsl_vector_set (thetaSubX, iter1, atof (buffer));
  } // for (iter1)
  inFile.close (); 
  inFile.clear ();
  

  /* tempVectors and the like store intermediate calculations: */
  gsl_vector *tempVector4 = gsl_vector_alloc (numberOfKnotBValues);
  gsl_vector *meanPartLengthKnotsPlusThree = gsl_vector_alloc (numberOfKnotValuesPlusThree);
  gsl_matrix *sampleFinalMatrix 
    = gsl_matrix_alloc (numberOfYValues, numberOfIterationsDividedByFive);
  gsl_vector *sampleMeans = gsl_vector_alloc (numberOfYValues);
  gsl_vector *sampleStandardDeviations = gsl_vector_alloc (numberOfYValues);
  gsl_vector *temp_1 = gsl_vector_alloc (numberOfKnotValuesPlusThree);
  gsl_matrix *temp_5 = gsl_matrix_alloc (numberOfThetaSubBValues, numberOfKnotValues);
  gsl_vector *temp_4 = gsl_vector_alloc (numberOfThetaSubBValues);
 

  /* Constants needed in algorithm: */
  double oneDividedByTwoSigma2U = 1.0 / (2.0 * sigma2U);
  double oneDividedBySigma2V = 1.0 / sigma2V;
  double aSubC = 1e-3;
  double aSubD = 1e-3;
  double bSubC = 1e-3;
  double bSubD = 1e-3;
  double sigma20Beta = 1e6;
  double sigma20Gamma = 1e6;
  double sigma20Delta = 1e6;
  int p = 2;
  int q = 1;
  int kSupEpsilon = numberOfKnotValues;
  int kSupB = numberOfKnotBValues;
  double gammaShapeParameter1 = aSubC + (kSupEpsilon / 2.0);
  double gammaShapeParameter2 = aSubD + (kSupB / 2.0);
  double reciprocalSigma2C = 0.0;
  double reciprocalSigma2D = 0.0;
  double mHProposal = 0.0;
  double a = 0.0;
  double b = 0.0;
  double c = 0.0;
  double d = 0.0;
  double e = 0.0;

 
  /* Vectors and matrices needed in algorithm: */
  gsl_vector *ySubEpsilon = gsl_vector_alloc (numberOfYValues);
  
  gsl_vector *ySubB = gsl_vector_alloc (numberOfKnotValues);

  gsl_vector *mu = gsl_vector_alloc (numberOfYValues);
  gsl_vector *eta = gsl_vector_alloc (numberOfYValues);
  gsl_vector *si = gsl_vector_alloc (numberOfKnotValues);

  gsl_matrix *cSubX = gsl_matrix_alloc (numberOfYValues, numberOfKnotValuesPlusThree);
  gsl_matrix *cSubXTranspose = gsl_matrix_alloc (numberOfKnotValuesPlusThree, numberOfYValues);
  gsl_matrix *cSubXTransposeSigmaSubEpsilonInverse 
    = gsl_matrix_alloc (numberOfKnotValuesPlusThree, numberOfYValues);

  gsl_matrix *cSubEpsilon = gsl_matrix_alloc (numberOfYValues, numberOfKnotValuesPlusThree);
  gsl_matrix *cSubEpsilonTransposeDividedBySigma2U 
    = gsl_matrix_alloc (numberOfKnotValuesPlusThree, numberOfYValues);
  gsl_matrix *cSubEpsilonTransposeCSubEpsilonDividedBySigma2U 
    = gsl_matrix_alloc (numberOfKnotValuesPlusThree, numberOfKnotValuesPlusThree);

  gsl_matrix *cSubB = gsl_matrix_alloc (numberOfKnotValues, numberOfThetaSubBValues);
  gsl_matrix *cSubBTransposeDividedBySigma2V = gsl_matrix_alloc (numberOfThetaSubBValues, numberOfKnotValues);
  gsl_matrix *cSubBTransposeCSubBDividedBySigma2V = gsl_matrix_alloc (numberOfThetaSubBValues, numberOfThetaSubBValues);

  /* sigmaSub0XInverse diagonal: */
  gsl_matrix *sigmaSub0XInverse = gsl_matrix_calloc (numberOfKnotValuesPlusThree, numberOfKnotValuesPlusThree);
  for (iter1 = 0; iter1 < p + 1; ++iter1) {
    gsl_matrix_set (sigmaSub0XInverse, iter1, iter1, 1.0 / sigma20Beta);
  } // for (iter1)

  /* sigmaSub0EpsilonInverse diagonal: */
  gsl_matrix *sigmaSub0EpsilonInverse = gsl_matrix_calloc (numberOfKnotValuesPlusThree, numberOfKnotValuesPlusThree);
  for (iter1 = 0; iter1 < p + 1; ++iter1) {
    gsl_matrix_set (sigmaSub0EpsilonInverse, iter1, iter1, 1.0 / sigma20Gamma);
  } // for (iter1) 

  /* sigmaSub0BInverse diagonal: */
  gsl_matrix *sigmaSub0BInverse = gsl_matrix_calloc (numberOfThetaSubBValues, numberOfThetaSubBValues);
  for (iter1 = 0; iter1 < q + 1; ++iter1) {
    gsl_matrix_set (sigmaSub0BInverse, iter1, iter1, 1.0 / sigma20Delta);
  } // for (iter1)

  gsl_matrix *mSubXInverse = gsl_matrix_calloc (numberOfKnotValuesPlusThree, numberOfKnotValuesPlusThree);
  gsl_matrix *mSubEpsilonInverse = gsl_matrix_alloc (numberOfKnotValuesPlusThree, numberOfKnotValuesPlusThree);
  gsl_matrix *mSubBInverse = gsl_matrix_alloc (numberOfThetaSubBValues, numberOfThetaSubBValues);


  /* Vector and matrices for saving samples: */
  gsl_vector *sampleSigma2C 
    = gsl_vector_alloc (numberOfIterationsDividedByFive);
  gsl_vector *sampleSigma2D 
    = gsl_vector_alloc (numberOfIterationsDividedByFive);
  gsl_matrix *sampleThetaSubX 
    = gsl_matrix_alloc (numberOfIterationsDividedByFive, numberOfKnotValuesPlusThree);
  gsl_matrix *sampleThetaSubEpsilon 
    = gsl_matrix_alloc (numberOfIterationsDividedByFive, numberOfKnotValuesPlusThree);
  gsl_matrix *sampleThetaSubB 
    = gsl_matrix_alloc (numberOfIterationsDividedByFive, numberOfThetaSubBValues);
  gsl_matrix *sampleSigma2Epsilon 
    = gsl_matrix_alloc (numberOfIterationsDividedByFive, numberOfYValues);
  gsl_matrix *sampleSigma2B 
    = gsl_matrix_alloc (numberOfIterationsDividedByFive, numberOfKnotValues);

  /* Using Tauslife pseudo-random number alogorithm,
   * the fastest simulation-quality algorithm provided by GSL: */
  gsl_rng *ranGen = gsl_rng_alloc (gsl_rng_taus);
  /* Seed with current time: */
  gsl_rng_set (ranGen, (unsigned long int)(time (NULL)));
  

  /* Calculate cSubX: */
  /* First column all ones: */
  gsl_vector_set_all (&(gsl_matrix_column (cSubX, 0)).vector, 1.0);
  /* Second column standardized range values in tempVectorLengthY: */
  gsl_matrix_set_col (cSubX, 1, tempVectorLengthY);
  /* Third column squared values of standardized range in tempVectorLengthY: */
  for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
    gsl_matrix_set (cSubX, iter1, 2, gsl_pow_2 (gsl_vector_get (tempVectorLengthY, iter1)));
  } // for (iter1)
  /* Rest of columns are squared difference of std. range and knots or 0: */
  for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
    for (iter2 = 0; iter2 < numberOfKnotValues; ++iter2) {
      gsl_matrix_set (cSubX, iter1, iter2 + 3,
		      gsl_pow_2 (max (gsl_vector_get (tempVectorLengthY, iter1)
			   - gsl_vector_get (tempVectorLengthKnots, iter2), 0.0)));
    } // for (iter2)
  } // for (iter1)


  /* Save transpose of cSubX for later use: */
  gsl_matrix_transpose_memcpy (cSubXTranspose, cSubX);


  /* Calculate cSubEpsilonTranspose variants: */
  /* Copy cSubX into cSubEpsilon: */
  gsl_matrix_memcpy (cSubEpsilon, cSubX);
  gsl_matrix_transpose_memcpy (cSubEpsilonTransposeDividedBySigma2U,
			       cSubEpsilon); 
  gsl_matrix_scale (cSubEpsilonTransposeDividedBySigma2U, 1.0 / sigma2U); 
  /* Multiply cSubEpsilonTranspose and cSubEpsilon, store in
   * cSubEpsilonTransposeCSubEpsilonDividedBySigma2U: */
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 
		  1.0, cSubEpsilonTransposeDividedBySigma2U, cSubEpsilon, 
		  0.0, cSubEpsilonTransposeCSubEpsilonDividedBySigma2U);

   
  /* Calculate cSubB: */
  /* First column ones: */
  gsl_vector_set_all (&(gsl_matrix_column (cSubB, 0)).vector, 1.0);
  /* Second column knot values: */
  gsl_matrix_set_col (cSubB, 1, tempVectorLengthKnots);
  for (iter1 = 0; iter1 < numberOfKnotValues; ++iter1) {
    for (iter2 = 0; iter2 < numberOfKnotBValues; ++iter2) {
      gsl_matrix_set (cSubB, iter1, iter2 + 2,
		      gsl_pow_2 (max (gsl_vector_get (tempVectorLengthKnots, iter1)
			   - gsl_vector_get (knotsB, iter2), 0.0)));
    } // for (iter2)
  } // for (iter1)
  /* Multiply cSubBTranspose by cSubB, store in
   * cSubBTransposeCSubBDividedBySigma2V: */
  gsl_matrix_transpose_memcpy (cSubBTransposeDividedBySigma2V, cSubB);
  gsl_matrix_scale (cSubBTransposeDividedBySigma2V, 1.0 / sigma2V);
  gsl_blas_dgemm (CblasNoTrans, CblasNoTrans,
		  1.0, cSubBTransposeDividedBySigma2V, cSubB,
		  0.0, cSubBTransposeCSubBDividedBySigma2V);



    /* Main loop: ************************** Start of main loop of algorithm *********/
    for (mainIter = 1; mainIter <= numberOfBurnInIterations
	   + numberOfIterations; ++mainIter) {
      cout << mainIter << endl;
      

      /* Steps 4 and 5: generate sigma2C and sigma2D. ********************************/
      /* Copy subvector for temporary use: */
      gsl_vector_memcpy (tempVectorLengthKnots, 
			 &(gsl_vector_subvector (thetaSubEpsilon, 
						 3, numberOfKnotValues)).vector);
      /* Take dot product of the subvector and store in tempDouble1: */
      gsl_blas_ddot (tempVectorLengthKnots, tempVectorLengthKnots, &tempDouble1);
      tempDouble1 /= 2.0;
      /* Sigma2C is an inverse Gamma, so reciprocalSigma2C is
       * a gamma random value: */
      reciprocalSigma2C = gsl_ran_gamma (ranGen, gammaShapeParameter1,
					 bSubC + tempDouble1);
      /* Repeat as above: */
      gsl_vector_memcpy (tempVector4, 
			 &(gsl_vector_subvector (thetaSubB, 2, 4)).vector);
      gsl_blas_ddot (tempVector4, tempVector4, &tempDouble1);
      tempDouble1 /= 2.0;
      reciprocalSigma2D = gsl_ran_gamma (ranGen, gammaShapeParameter2,
					 bSubD + tempDouble1);


      
      /* Calculate vectors mu, eta, and si for M-H steps: */
      /* The last argument to the blas commands is overwritten. */
      gsl_blas_dgemv (CblasNoTrans,
		      1.0, cSubX, thetaSubX,
		      0.0, mu);
      gsl_blas_dgemv (CblasNoTrans,
		      1.0, cSubEpsilon, thetaSubEpsilon,
		      0.0, eta);
      gsl_blas_dgemv (CblasNoTrans,
		      1.0, cSubB, thetaSubB,
		      0.0, si);


      /* Step 6: M-H. ****************************************************************/
      /* Calculate y_i - mu_i for all i = 0,...,numberOfYValues and
       * store in tempVectorLengthY */
      /* Begin by storing the y values in tempVectorLengthY: */
      gsl_vector_memcpy (tempVectorLengthY, y);
      /* Subtract mu values from y values element-wise: */
      gsl_vector_sub (tempVectorLengthY, mu);
      
      double acceptRatio = 0.0;
      
      /* The M-H step: */
      for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
	a = gsl_vector_get (sigma2Epsilon, iter1);
	c = gsl_vector_get (eta, iter1);
	d = gsl_vector_get (y, iter1);
	e = gsl_vector_get (mu, iter1);

//  	for (iter2 = 0; iter2 < 5; ++iter2) {
	mHProposal = a + gsl_ran_gaussian (ranGen, sqrt (proposalVariance));
	if (mHProposal > 0.0) {
	  b = pow (mHProposal / a, -1.5) 
	    * exp ((gsl_pow_2 (d - e)/2.) * (1 / a - 1 / mHProposal));
	  b *= exp (oneDividedByTwoSigma2U 
		    * (gsl_pow_2 (log (a) - c) 
		       - gsl_pow_2 (log (mHProposal) - c)));
	  if (b >= 1 || gsl_rng_uniform (ranGen) < b) {
	    a = mHProposal;
 	    ++acceptRatio;
	  } // if (b)
	} // if (mHProposal)
//        } // for (iter2)
	gsl_vector_set (sigma2Epsilon, iter1, a);
      } // for (iter1)
//       if (mainIter == 1) {
// 	f.open ("acceptanceRatioStep6.txt", ios::out | ios::trunc);
// 	f.close ();
//       }
//       f.open ("acceptanceRatioStep6.txt", ios::out | ios::app);
//       f << "Acceptance ratio: " 
//  	   << acceptRatio / numberOfYValues << endl;
//       f.close ();

        acceptRatio = 0.0;

      /* Step 7: M-H: ****************************************************/
      for (iter1 = 0; iter1 < numberOfKnotValues; ++iter1) {
	a = gsl_vector_get (sigma2B, iter1);
	c = gsl_vector_get (si, iter1);
	d = gsl_vector_get (thetaSubX, iter1 + 3);

// 	for (iter2 = 0; iter2 < 5; ++iter2) {
	mHProposal = a + gsl_ran_gaussian (ranGen, proposalVariance);
	  if (mHProposal > 0.0) {
	    b = pow (mHProposal / a, -1.5);
	    b *= exp((((gsl_pow_2(d))/2.0)*(1/a-1/mHProposal))
		     +((oneDividedBySigma2V)
		       *(gsl_pow_2(log(a)-c)-gsl_pow_2(log(mHProposal)-c))));
	    if (b >= 1.0 || gsl_rng_uniform (ranGen) < b) {
	      a = mHProposal;
 	      ++acceptRatio;
	    } // if (b)
	  } // if (mHProposal)
// 	} // for (iter2)
	gsl_vector_set (sigma2B, iter1, a);
      } // for (iter1)
//       if (mainIter == 1) {
// 	f.open ("acceptanceRatioStep7.txt", ios::out | ios::trunc);
// 	f.close ();
//       }
//       f.open ("acceptanceRatioStep7.txt", ios::out | ios::app);
//       f << "Acceptance ratio for Step 7: " 
// 	<< acceptRatio / numberOfKnotValues << endl;
//       f.close ();
      

      /* Create the matrices to be added in creating mSub_: */
      /* sigmaSub0XInverse, which is diagonal,
       * consists of reciprocal of sigma2B values
       * for the first few diagonal elements, and the
       * inverse Gamma values for the rest of the diagonal: */
      for (iter1 = 0; iter1 < kSupEpsilon; ++iter1) {
	gsl_matrix_set (sigmaSub0XInverse,
			iter1 + p + 1,
			iter1 + p + 1,
			1.0 / gsl_vector_get (sigma2B, iter1));
	gsl_matrix_set (sigmaSub0EpsilonInverse,
			iter1 + p + 1,
			iter1 + p + 1,
			reciprocalSigma2C);
      } // for (iter1)
      /* sigmaSub0BInverse is diagonal consisting of
       * inverse Gamma values: */
      for (iter1 = 0; iter1 < kSupB; ++iter1) {
	gsl_matrix_set (sigmaSub0BInverse,
			iter1 + q + 1,
			iter1 + q + 1,
			reciprocalSigma2D);
      } // for (iter1)


      /* Step 1: **********************************************************/
      /* Preparation for calculating mSubXInverse:
       * cSubXTranspose * sigmaSubEpsilonInverse;
       * here, sigma2Epsilon is the vector of diagonal elements of
       * sigmaSubEpsilon */
      gsl_matrix_memcpy (cSubXTransposeSigmaSubEpsilonInverse, 
			 cSubXTranspose);
      multiplyByDiagonalInverseOnRight (cSubXTransposeSigmaSubEpsilonInverse,
					sigma2Epsilon);
     

       /* mSubXInverse: */
       /* cSubXTranspose * sigmaSubEpsilonInverse * cSubX 
	* + sigmaSub0XInverse */
       /* Copy sigmaSub0XInverse into mSubXInverse to avoid
	* resetting sigmaSub0XInverse every iteration: */

       
       gsl_matrix_memcpy (mSubXInverse, sigmaSub0XInverse);
       gsl_blas_dgemm (CblasNoTrans, CblasNoTrans,
		       1.0, cSubXTransposeSigmaSubEpsilonInverse, cSubX, 
		       1.0, mSubXInverse);
       /* cSubXTranspose * sigmaSubEpsilonInverse * Y, in prepartion for
	* generating realization from multivariate normal density: */
       /* tempVectorLengthY is overwritten: */
       gsl_blas_dgemv (CblasNoTrans,
		       1.0, cSubXTransposeSigmaSubEpsilonInverse, y,
		       0.0, meanPartLengthKnotsPlusThree);

       /* Generate thetaSubX: */
      
       
       
       /* See notes at function definition at bottom of file. */
       multinvpdsymm_left (mSubXInverse, meanPartLengthKnotsPlusThree, temp_1);
 
       /* Make vector of random std. normal values: */
       for (iter1 = 0; iter1 < numberOfKnotValuesPlusThree; ++iter1) {
	 gsl_vector_set (thetaSubX, iter1, gsl_ran_gaussian (ranGen, 1.0));
       } // for (iter1)

       /* mSubXInverse has been overwritten with its Cholesky decomposition. */
       multinvupper_left (mSubXInverse, thetaSubX);
       /* thetaSubX has been overwritten with inverse(L') * ran_gaussian_vector. */
       gsl_vector_add (temp_1, thetaSubX);
       gsl_vector_memcpy (thetaSubX, temp_1);



       /* Step 2: *****************************************************************/
       /* mSubEpsilonInverse: */
       /* Overwrite sigmaSub0EpsilonIverse. */
       gsl_matrix_memcpy (mSubEpsilonInverse, sigmaSub0EpsilonInverse);
       gsl_matrix_add (mSubEpsilonInverse,
		       cSubEpsilonTransposeCSubEpsilonDividedBySigma2U);
       /* cSubEpsilonTranspose * ySubEpsilon / sigma2U, part of multivariate
	* normal mean: */
       /* ySubEpsilon is log(sigma2Epsilon): */
       for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
	 gsl_vector_set (ySubEpsilon, iter1, 
			 log (gsl_vector_get (sigma2Epsilon, iter1)));
       } // for (iter1)
       gsl_blas_dgemv (CblasNoTrans,
		       1.0, cSubEpsilonTransposeDividedBySigma2U,
		       ySubEpsilon,
		       0.0, meanPartLengthKnotsPlusThree);

       /* Generate thetaSubEpsilon: */

       
       /* See notes at function definition at bottom of file. */
       multinvpdsymm_left (mSubEpsilonInverse, meanPartLengthKnotsPlusThree, temp_1);

       for (iter1 = 0; iter1 < numberOfKnotValuesPlusThree; ++iter1) {
	 gsl_vector_set (thetaSubEpsilon, iter1, gsl_ran_gaussian (ranGen, 1.0));
       } // for (iter1)

       /* mSubEpsilonInverse has been overwritten with its Cholesky decomposition. */
       multinvupper_left (mSubEpsilonInverse, thetaSubEpsilon);
       /* thetaSubEpsilon has been overwritten with inverse(L') * ran_gaussian_vector. */
       gsl_vector_add (temp_1, thetaSubEpsilon);
       gsl_vector_memcpy (thetaSubEpsilon, temp_1);


       /* Step 3: *********************************************************************/
       /* mSubBInverse: */
       /* Overwrite sigmaSub0BInverse. */
       gsl_matrix_memcpy (mSubBInverse, sigmaSub0BInverse);
       gsl_matrix_add (mSubBInverse,
		       cSubBTransposeCSubBDividedBySigma2V);
       /* ySubB is log(sigma2B): */
       for (iter1 = 0; iter1 < numberOfKnotValues; ++iter1) {
	 gsl_vector_set (ySubB, iter1, 
			 log (gsl_vector_get (sigma2B, iter1)));
       } // for (iter1)


       /* See notes at function definition at bottom of file. */
       multinvpdsymm_left (mSubBInverse, cSubBTransposeDividedBySigma2V, temp_5);

       /* Product of solution to the Cholesky system above and ySubB: */
       gsl_blas_dgemv (CblasNoTrans,
		       1.0, temp_5, ySubB,
		       0.0, temp_4);

       for (iter1 = 0; iter1 < numberOfThetaSubBValues; ++iter1) {
	 gsl_vector_set (thetaSubB, iter1, gsl_ran_gaussian (ranGen, 1.0));
       } // for (iter1)

       /* mSubBInverse has been overwritten with its Cholesky decomposition. */
       multinvupper_left (mSubBInverse, thetaSubB);
       /* thetaSubB has been overwritten with inverse(L') * ran_gaussian_vector. */
       gsl_vector_add (temp_4, thetaSubB);
       gsl_vector_memcpy (thetaSubB, temp_4);

      
       /* Save samples every 5th iteration past numberOfBurnInIterations: */
       if ((mainIter > numberOfBurnInIterations) && (mainIter % 5 == 0)) {
	 gsl_matrix_set_row (sampleThetaSubX, saveRow, thetaSubX);
	 gsl_matrix_set_row (sampleThetaSubEpsilon, saveRow, thetaSubEpsilon);
	 gsl_matrix_set_row (sampleThetaSubB, saveRow, thetaSubB);
	 gsl_vector_set (sampleSigma2C, saveRow, 1.0 / reciprocalSigma2C);
	 gsl_vector_set (sampleSigma2D, saveRow, 1.0 / reciprocalSigma2D);
	 gsl_matrix_set_row (sampleSigma2Epsilon, saveRow, sigma2Epsilon);
	 gsl_matrix_set_row (sampleSigma2B, saveRow, sigma2B);

	 ++saveRow;
       } // if (mainIter)
    } // for (mainIter) *************** End of main loop of algorithm **********


    /* Print out samples: */
    f.open ("sampleThetaSubX.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
      for (int j = 0; j < numberOfKnotValuesPlusThree; ++j) {
	f << fixed << gsl_matrix_get (sampleThetaSubX, i, j) << " ";
      } // for (j)
      f << endl;
    } // for (i)
    f.close ();
    f.open ("sampleThetaSubEpsilon.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
      for (int j = 0; j < numberOfKnotValuesPlusThree; ++j) {
	f << fixed << gsl_matrix_get (sampleThetaSubEpsilon, i, j) << " ";
      } // for (j)
      f << endl;
    } // for (i)
    f.close ();
    f.open ("sampleThetaSubB.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
      for (int j = 0; j < numberOfThetaSubBValues; ++j) {
	f << fixed << gsl_matrix_get (sampleThetaSubB, i, j) << " ";
      } // for (j)
      f << endl;
    } // for (i)
    f.close ();
    f.open ("sampleSigma2C.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
	f << fixed << gsl_vector_get (sampleSigma2C, i)
	  << endl;
    } // for (i)
    f.close ();
     f.open ("sampleSigma2D.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
      f << fixed << gsl_vector_get (sampleSigma2D, i)
	<< endl;
    } // for (i)
    f.close ();
    f.open ("sampleSigma2Epsilon.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
      for (int j = 0; j < numberOfYValues; ++j) {
	f << fixed << gsl_matrix_get (sampleSigma2Epsilon, i, j) << " ";
      } // for (j)
      f << endl;
    } // for (i)
    f.close ();
    f.open ("sampleSigma2B.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfIterationsDividedByFive; ++i) {
      for (int j = 0; j < numberOfKnotValues; ++j) {
	f << fixed << gsl_matrix_get (sampleSigma2B, i, j) << " ";
      } // for (j)
      f << endl;
    } // for (i)
    f.close ();
    

    /* Calculate and write sample means and sample standard deviations of posterior: */
    gsl_blas_dgemm (CblasNoTrans, CblasTrans,
		    1.0, cSubX, sampleThetaSubX,
		    0.0, sampleFinalMatrix);

    /* Calculate sample means: */
    tempDouble1 = 1.0 / (double)(numberOfIterationsDividedByFive);
    for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
      tempDouble2 = 0.0;
      for (iter2 = 0; iter2 < numberOfIterationsDividedByFive; ++iter2) {
	tempDouble2 += gsl_matrix_get (sampleFinalMatrix, iter1, iter2);
      } // for (iter2)
      gsl_vector_set (sampleMeans, iter1, tempDouble1 * tempDouble2);
    } // for (iter1)

    /* Calculate sample standard deviations: */
    tempDouble1 = 1.0 / (double)(numberOfIterationsDividedByFive - 1);
    for (iter1 = 0; iter1 < numberOfYValues; ++iter1) {
      tempDouble2 = 0.0;
      for (iter2 = 0; iter2 < numberOfIterationsDividedByFive; ++iter2) {
	tempDouble2 += 
	  sqrt (gsl_pow_2 (gsl_matrix_get (sampleFinalMatrix, iter1, iter2)
	   - gsl_vector_get (sampleMeans, iter1)));
      } // for (iter2)
      gsl_vector_set (sampleStandardDeviations, iter1, tempDouble1 * tempDouble2);
    } // for (iter1)

    /* Write to file: */
    f.open ("gslSampleMeansAndStandardDeviations.txt", ios::out | ios::trunc);
    for (int i = 0; i < numberOfYValues; ++i) {
      f << fixed << gsl_vector_get (sampleMeans, i) 
	<< " " << gsl_vector_get (sampleStandardDeviations, i) << endl;
    } // for (i)
    f.close ();
 

  /* Print total execution time to screen. */
  cout << "Run time: "
       << difftime (time (NULL), runTimeStart)
       << " s.\n";

  return 0;
} // main ()



/* Compute the product of A and the inverse of a square diagonal matrix.
 * input:   matrix A
 *          vector Diag, which represents the diagonal elements of
 *            a square matrix
 * output:  overwrites A with the product of A and the inverse of
 *            the square diagonal matrix
 */
void multiplyByDiagonalInverseOnRight (gsl_matrix * A, 
				       gsl_vector * Diag) {
  int colsA = (int)(*A).size2;
  for (int i = 0; i < colsA; ++i) {
    /* Set i_th column of A to Diag_i * i_th column: */
    gsl_vector_scale (&(gsl_matrix_column (A, i)).vector, 
		      1.0 / gsl_vector_get (Diag, i));
  }// for (i)
} // multiplyByDiagonalInverseOnRight ()

/* Compute Ax = B where A is symmetric positive definite and B is a vector.
 *
 * input:   symmetric positive definite matrix A
 *          vector B
 *
 * output:  overwrites A with the Cholesky decomposition LL' with L stored
 *            in the lower part of A and L' stored in the upper part of A
 *            with shared diagonal elements.
 *          overwrites C with the solution to the system
 */
void multinvpdsymm_left (gsl_matrix *A, gsl_vector *B, gsl_vector *C) {
  gsl_linalg_cholesky_decomp (A);
  gsl_linalg_cholesky_solve (A, B, C);
} // multinvpdsymm_left ()

/* Same as above except B is now a matrix. */
void multinvpdsymm_left (gsl_matrix *A, gsl_matrix *B, gsl_matrix *C) {
  int n = (int)(*C).size2;
  gsl_linalg_cholesky_decomp (A);
  for (int j = 0; j < n; ++j) {
    gsl_linalg_cholesky_solve (A, &(gsl_matrix_column (B, j)).vector, &(gsl_matrix_column (C, j)).vector);
  } // for (j)
} // multinvpdsymm_left ()

/* Computes product of inverse(U) and A, overwritten
 *
 * input:   matrix U of which the upper triangular part will be used
 *          vector A
 *          
 * output:  overwrites A with the product of inverse(U) and A
 */
void multinvupper_left (gsl_matrix *U, gsl_vector *A) {
  int n = (int)(*A).size;
  double tempDouble = 0.0;
  double tempDouble2 = 0.0;
  double tempDouble3 = 0.0;
  gsl_vector *tempVector1;
  gsl_vector *tempVector2;
  gsl_vector *temp = gsl_vector_alloc ((int)(*U).size2);
  tempDouble2 = gsl_vector_get (A, n - 1);
  tempDouble3 = gsl_matrix_get (U, n - 1, n - 1);
  tempDouble = tempDouble2 / tempDouble3;
  gsl_vector_set (A, n - 1, tempDouble);
  for (int i = n - 2; i >= 0; --i) {
    gsl_matrix_get_row (temp, U, i);
    tempVector1 = gsl_vector_alloc (n - i - 1);
    tempVector2 = gsl_vector_alloc (n - i - 1);
    gsl_vector_memcpy (tempVector1, &(gsl_vector_subvector (temp, i + 1, n - i - 1)).vector);
    gsl_vector_memcpy (tempVector2, &(gsl_vector_subvector (A, i + 1, n - i - 1)).vector);
    gsl_blas_ddot (tempVector1, tempVector2, &tempDouble);
    gsl_vector_set (A, i,
		    (gsl_vector_get (A, i)
		    - tempDouble) / gsl_matrix_get (U, i, i));
    gsl_vector_free (tempVector1);
    gsl_vector_free (tempVector2);
  } // for (i)
    gsl_vector_free (temp);
} // multinvupper_left ()
